@@ -1,5 +1,5 @@
 # frozen_string_literal: true
-# $Id: test_verbose.rb 57275 2017-01-07 02:14:07Z kazu $
+# $Id$
 
 require 'test/unit'
 require 'fileutils'
@@ -8,7 +8,7 @@ require_relative 'visibility_tests'
 class TestFileUtilsVerbose < Test::Unit::TestCase
 
   include FileUtils::Verbose
-  include TestFileUtils::Visibility
+  include TestFileUtilsIncVisibility
 
   def setup
     super
