@@ -34,7 +34,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_class_Proxy_from_ENV
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy']      = 'http://proxy.example:8000'
 
       # These are ignored on purpose.  See Bug 4388 and Feature 6546
@@ -115,7 +115,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_address
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       http = Net::HTTP.new 'hostname.example', nil, 'proxy.example'
       assert_equal 'proxy.example', http.proxy_address
 
@@ -125,17 +125,17 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_address_no_proxy
-    clean_http_proxy_env do
-      http = Net::HTTP.new 'hostname.example', nil, 'proxy.example', nil, nil, nil, 'example'
+    TestNetHTTPUtils.clean_http_proxy_env do
+      http = Net::HTTP.new 'hostname.example', nil, 'proxy.com', nil, nil, nil, 'example'
       assert_nil http.proxy_address
 
-      http = Net::HTTP.new '10.224.1.1', nil, 'proxy.example', nil, nil, nil, 'example,10.224.0.0/22'
+      http = Net::HTTP.new '10.224.1.1', nil, 'proxy.com', nil, nil, nil, 'example,10.224.0.0/22'
       assert_nil http.proxy_address
     end
   end
 
   def test_proxy_from_env_ENV
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
       assert_equal false, Net::HTTP.proxy_class?
@@ -146,7 +146,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_address_ENV
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
       http = Net::HTTP.new 'hostname.example'
@@ -156,13 +156,13 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_eh_no_proxy
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       assert_equal false, Net::HTTP.new('hostname.example', nil, nil).proxy?
     end
   end
 
   def test_proxy_eh_ENV
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
       http = Net::HTTP.new 'hostname.example'
@@ -172,30 +172,37 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_eh_ENV_with_user
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://foo:bar@proxy.example:8000'
 
       http = Net::HTTP.new 'hostname.example'
 
       assert_equal true, http.proxy?
-      if Net::HTTP::ENVIRONMENT_VARIABLE_IS_MULTIUSER_SAFE
-        assert_equal 'foo', http.proxy_user
-        assert_equal 'bar', http.proxy_pass
-      else
-        assert_nil http.proxy_user
-        assert_nil http.proxy_pass
-      end
+      assert_equal 'foo', http.proxy_user
+      assert_equal 'bar', http.proxy_pass
+    end
+  end
+
+  def test_proxy_eh_ENV_with_urlencoded_user
+    TestNetHTTPUtils.clean_http_proxy_env do
+      ENV['http_proxy'] = 'http://Y%5CX:R%25S%5D%20%3FX@proxy.example:8000'
+
+      http = Net::HTTP.new 'hostname.example'
+
+      assert_equal true, http.proxy?
+      assert_equal "Y\\X", http.proxy_user
+      assert_equal "R%S] ?X", http.proxy_pass
     end
   end
 
   def test_proxy_eh_ENV_none_set
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       assert_equal false, Net::HTTP.new('hostname.example').proxy?
     end
   end
 
   def test_proxy_eh_ENV_no_proxy
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
       ENV['no_proxy']   = 'hostname.example'
 
@@ -204,7 +211,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_port
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       http = Net::HTTP.new 'example', nil, 'proxy.example'
       assert_equal 'proxy.example', http.proxy_address
       assert_equal 80, http.proxy_port
@@ -216,7 +223,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_proxy_port_ENV
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
       http = Net::HTTP.new 'hostname.example'
@@ -226,7 +233,7 @@ class TestNetHTTP < Test::Unit::TestCase
   end
 
   def test_newobj
-    clean_http_proxy_env do
+    TestNetHTTPUtils.clean_http_proxy_env do
       ENV['http_proxy'] = 'http://proxy.example:8000'
 
       http = Net::HTTP.newobj 'hostname.example'
@@ -235,25 +242,6 @@ class TestNetHTTP < Test::Unit::TestCase
     end
   end
 
-  def clean_http_proxy_env
-    orig = {
-      'http_proxy'      => ENV['http_proxy'],
-      'http_proxy_user' => ENV['http_proxy_user'],
-      'http_proxy_pass' => ENV['http_proxy_pass'],
-      'no_proxy'        => ENV['no_proxy'],
-    }
-
-    orig.each_key do |key|
-      ENV.delete key
-    end
-
-    yield
-  ensure
-    orig.each do |key, value|
-      ENV[key] = value
-    end
-  end
-
   def test_failure_message_includes_failed_domain_and_port
     # hostname to be included in the error message
     host = Struct.new(:to_s).new("<example>")
@@ -262,10 +250,22 @@ class TestNetHTTP < Test::Unit::TestCase
     def host.to_str; raise SocketError, "open failure"; end
     uri = Struct.new(:scheme, :hostname, :port).new("http", host, port)
     assert_raise_with_message(SocketError, /#{host}:#{port}/) do
-      clean_http_proxy_env{ Net::HTTP.get(uri) }
+      TestNetHTTPUtils.clean_http_proxy_env{ Net::HTTP.get(uri) }
     end
   end
 
+  def test_default_configuration
+    Net::HTTP.default_configuration = { open_timeout: 5 }
+    http = Net::HTTP.new 'hostname.example'
+    assert_equal 5, http.open_timeout
+    assert_equal 60, http.read_timeout
+
+    http.open_timeout = 10
+    assert_equal 10, http.open_timeout
+  ensure
+    Net::HTTP.default_configuration = nil
+  end
+
 end
 
 module TestNetHTTP_version_1_1_methods
@@ -302,6 +302,27 @@ module TestNetHTTP_version_1_1_methods
   def test_s_get
     assert_equal $test_net_http_data,
         Net::HTTP.get(config('host'), '/', config('port'))
+
+    assert_equal $test_net_http_data, Net::HTTP.get(
+      URI.parse("http://#{config('host')}:#{config('port')}")
+    )
+    assert_equal $test_net_http_data, Net::HTTP.get(
+      URI.parse("http://#{config('host')}:#{config('port')}"), "Accept" => "text/plain"
+    )
+  end
+
+  def test_s_get_response
+    res = Net::HTTP.get_response(
+      URI.parse("http://#{config('host')}:#{config('port')}")
+    )
+    assert_equal "application/octet-stream", res["Content-Type"]
+    assert_equal $test_net_http_data, res.body
+
+    res = Net::HTTP.get_response(
+      URI.parse("http://#{config('host')}:#{config('port')}"), "Accept" => "text/plain"
+    )
+    assert_equal "text/plain", res["Content-Type"]
+    assert_equal $test_net_http_data, res.body
   end
 
   def test_head
@@ -433,7 +454,11 @@ module TestNetHTTP_version_1_1_methods
   def test_post
     start {|http|
       _test_post__base http
+    }
+    start {|http|
       _test_post__file http
+    }
+    start {|http|
       _test_post__no_data http
     }
   end
@@ -469,9 +494,11 @@ module TestNetHTTP_version_1_1_methods
 
   def test_s_post
     url = "http://#{config('host')}:#{config('port')}/?q=a"
-    res = Net::HTTP.post(
+    res = assert_warning(/Content-Type did not set/) do
+      Net::HTTP.post(
               URI.parse(url),
               "a=x")
+    end
     assert_equal "application/x-www-form-urlencoded", res["Content-Type"]
     assert_equal "a=x", res.body
     assert_equal url, res["X-request-uri"]
@@ -529,6 +556,59 @@ module TestNetHTTP_version_1_1_methods
     assert_equal data, res.entity
   end
 
+  def test_timeout_during_HTTP_session_write
+    th = nil
+    # listen for connections... but deliberately do not read
+    TCPServer.open('localhost', 0) {|server|
+      port = server.addr[1]
+
+      conn = Net::HTTP.new('localhost', port)
+      conn.write_timeout = EnvUtil.apply_timeout_scale(0.01)
+      conn.read_timeout = EnvUtil.apply_timeout_scale(0.01) if windows?
+      conn.open_timeout = EnvUtil.apply_timeout_scale(1)
+
+      th = Thread.new do
+        err = !windows? ? Net::WriteTimeout : Net::ReadTimeout
+        assert_raise(err) do
+          assert_warning(/Content-Type did not set/) do
+            conn.post('/', "a"*50_000_000)
+          end
+        end
+      end
+      assert th.join(EnvUtil.apply_timeout_scale(10))
+    }
+  ensure
+    th&.kill
+    th&.join
+  end
+
+  def test_timeout_during_non_chunked_streamed_HTTP_session_write
+    th = nil
+    # listen for connections... but deliberately do not read
+    TCPServer.open('localhost', 0) {|server|
+      port = server.addr[1]
+
+      conn = Net::HTTP.new('localhost', port)
+      conn.write_timeout = EnvUtil.apply_timeout_scale(0.01)
+      conn.read_timeout = EnvUtil.apply_timeout_scale(0.01) if windows?
+      conn.open_timeout = EnvUtil.apply_timeout_scale(1)
+
+      req = Net::HTTP::Post.new('/')
+      data = "a"*50_000_000
+      req.content_length = data.size
+      req['Content-Type'] = 'application/x-www-form-urlencoded'
+      req.body_stream = StringIO.new(data)
+
+      th = Thread.new do
+        assert_raise(Net::WriteTimeout) { conn.request(req) }
+      end
+      assert th.join(10)
+    }
+  ensure
+    th&.kill
+    th&.join
+  end
+
   def test_timeout_during_HTTP_session
     bug4246 = "expected the HTTP session to have timed out but have not. c.f. [ruby-core:34203]"
 
@@ -538,15 +618,15 @@ module TestNetHTTP_version_1_1_methods
       port = server.addr[1]
 
       conn = Net::HTTP.new('localhost', port)
-      conn.read_timeout = 0.01
-      conn.open_timeout = 0.1
+      conn.read_timeout = EnvUtil.apply_timeout_scale(0.01)
+      conn.open_timeout = EnvUtil.apply_timeout_scale(1)
 
       th = Thread.new do
         assert_raise(Net::ReadTimeout) {
           conn.get('/')
         }
       end
-      assert th.join(10), bug4246
+      assert th.join(EnvUtil.apply_timeout_scale(10)), bug4246
     }
   ensure
     th.kill
@@ -565,10 +645,12 @@ module TestNetHTTP_version_1_2_methods
       # _test_request__range http   # WEBrick does not support Range: header.
       _test_request__HEAD http
       _test_request__POST http
-      _test_request__stream_body http
       _test_request__uri http
       _test_request__uri_host http
     }
+    start {|http|
+      _test_request__stream_body http
+    }
   end
 
   def _test_request__GET(http)
@@ -779,7 +861,13 @@ Content-Type: application/octet-stream
 __EOM__
       start {|http|
         _test_set_form_urlencoded(http, data.reject{|k,v|!v.is_a?(String)})
+      }
+      start {|http|
+        @server.mount('/', lambda {|req, res| res.body = req.body })
         _test_set_form_multipart(http, false, data, expected)
+      }
+      start {|http|
+        @server.mount('/', lambda {|req, res| res.body = req.body })
         _test_set_form_multipart(http, true, data, expected)
       }
     }
@@ -823,6 +911,7 @@ __EOM__
       expected.sub!(/<filename>/, filename)
       expected.sub!(/<data>/, $test_net_http_data)
       start {|http|
+        @server.mount('/', lambda {|req, res| res.body = req.body })
         data.each{|k,v|v.rewind rescue nil}
         req = Net::HTTP::Post.new('/')
         req.set_form(data, 'multipart/form-data')
@@ -838,10 +927,11 @@ __EOM__
           header)
         assert_equal(expected, body)
 
-        data.each{|k,v|v.rewind rescue nil}
-        req['Transfer-Encoding'] = 'chunked'
-        res = http.request req
-        #assert_equal(expected, res.body)
+        # TODO: test with chunked
+        # data.each{|k,v|v.rewind rescue nil}
+        # req['Transfer-Encoding'] = 'chunked'
+        # res = http.request req
+        # assert_equal(expected, res.body)
       }
     }
   end
@@ -862,6 +952,17 @@ class TestNetHTTP_v1_2 < Test::Unit::TestCase
     Net::HTTP.version_1_2
     super
   end
+
+  def test_send_large_POST_request
+    start {|http|
+      data = ' '*6000000
+      res = http.send_request('POST', '/', data, 'content-type' => 'application/x-www-form-urlencoded')
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+      assert_equal data.size, res.body.size
+      assert_equal data, res.body
+    }
+  end
 end
 
 class TestNetHTTP_v1_2_chunked < Test::Unit::TestCase
@@ -909,7 +1010,7 @@ class TestNetHTTPContinue < Test::Unit::TestCase
   end
 
   def mount_proc(&block)
-    @server.mount('/continue', WEBrick::HTTPServlet::ProcHandler.new(block.to_proc))
+    @server.mount('/continue', block.to_proc)
   end
 
   def test_expect_continue
@@ -964,10 +1065,10 @@ class TestNetHTTPContinue < Test::Unit::TestCase
   def test_expect_continue_error_before_body
     @log_tester = nil
     mount_proc {|req, res|
-      raise WEBrick::HTTPStatus::Forbidden
+      raise TestNetHTTPUtils::Forbidden
     }
     start {|http|
-      uheader = {'content-length' => '5', 'expect' => '100-continue'}
+      uheader = {'content-type' => 'application/x-www-form-urlencoded', 'content-length' => '5', 'expect' => '100-continue'}
       http.continue_timeout = 1 # allow the server to respond before sending
       http.request_post('/continue', 'data', uheader) {|res|
         assert_equal(res.code, '403')
@@ -1009,7 +1110,7 @@ class TestNetHTTPSwitchingProtocols < Test::Unit::TestCase
   end
 
   def mount_proc(&block)
-    @server.mount('/continue', WEBrick::HTTPServlet::ProcHandler.new(block.to_proc))
+    @server.mount('/continue', block.to_proc)
   end
 
   def test_info
@@ -1019,7 +1120,8 @@ class TestNetHTTPSwitchingProtocols < Test::Unit::TestCase
     }
     start {|http|
       http.continue_timeout = 0.2
-      http.request_post('/continue', 'body=BODY') {|res|
+      http.request_post('/continue', 'body=BODY',
+                        'content-type' => 'application/x-www-form-urlencoded') {|res|
         assert_equal('BODY', res.read_body)
       }
     }
@@ -1082,6 +1184,30 @@ class TestNetHTTPKeepAlive < Test::Unit::TestCase
     }
   end
 
+  def test_keep_alive_reset_on_new_connection
+    # Using debug log output on accepting connection:
+    #
+    #   "[2021-04-29 20:36:46] DEBUG accept: 127.0.0.1:50674\n"
+    @log_tester = nil
+    @logger_level = :debug
+
+    start {|http|
+      res = http.get('/')
+      http.keep_alive_timeout = 1
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+      http.finish
+      assert_equal 1, @log.grep(/accept/i).size
+
+      sleep 1.5
+      http.start
+      res = http.get('/')
+      assert_kind_of Net::HTTPResponse, res
+      assert_kind_of String, res.body
+      assert_equal 2, @log.grep(/accept/i).size
+    }
+  end
+
   class MockSocket
     attr_reader :count
     def initialize(success_after: nil)
@@ -1134,6 +1260,16 @@ class TestNetHTTPKeepAlive < Test::Unit::TestCase
     }
   end
 
+  def test_http_retry_failed_with_block
+    start {|http|
+      http.max_retries = 10
+      called = 0
+      assert_raise(Errno::ECONNRESET){ http.get('/'){called += 1; raise Errno::ECONNRESET} }
+      assert_equal 1, called
+    }
+    @log_tester = nil
+  end
+
   def test_keep_alive_server_close
     def @server.run(sock)
       sock.close
@@ -1184,3 +1320,87 @@ class TestNetHTTPLocalBind < Test::Unit::TestCase
   end
 end
 
+class TestNetHTTPForceEncoding < Test::Unit::TestCase
+  CONFIG = {
+    'host' => 'localhost',
+    'proxy_host' => nil,
+    'proxy_port' => nil,
+  }
+
+  include TestNetHTTPUtils
+
+  def fe_request(force_enc, content_type=nil)
+    @server.mount_proc('/fe') do |req, res|
+      res['Content-Type'] = content_type if content_type
+      res.body = "hello\u1234"
+    end
+
+    http = Net::HTTP.new(config('host'), config('port'))
+    http.local_host = Addrinfo.tcp(config('host'), config('port')).ip_address
+    assert_not_nil(http.local_host)
+    assert_nil(http.local_port)
+
+    http.response_body_encoding = force_enc
+    http.get('/fe')
+  end
+
+  def test_response_body_encoding_false
+    res = fe_request(false)
+    assert_equal("hello\u1234".b, res.body)
+    assert_equal(Encoding::ASCII_8BIT, res.body.encoding)
+  end
+
+  def test_response_body_encoding_true_without_content_type
+    res = fe_request(true)
+    assert_equal("hello\u1234".b, res.body)
+    assert_equal(Encoding::ASCII_8BIT, res.body.encoding)
+  end
+
+  def test_response_body_encoding_true_with_content_type
+    res = fe_request(true, 'text/html; charset=utf-8')
+    assert_equal("hello\u1234", res.body)
+    assert_equal(Encoding::UTF_8, res.body.encoding)
+  end
+
+  def test_response_body_encoding_string_without_content_type
+    res = fe_request('utf-8')
+    assert_equal("hello\u1234", res.body)
+    assert_equal(Encoding::UTF_8, res.body.encoding)
+  end
+
+  def test_response_body_encoding_encoding_without_content_type
+    res = fe_request(Encoding::UTF_8)
+    assert_equal("hello\u1234", res.body)
+    assert_equal(Encoding::UTF_8, res.body.encoding)
+  end
+end
+
+class TestNetHTTPPartialResponse < Test::Unit::TestCase
+  CONFIG = {
+    'host' => '127.0.0.1',
+    'proxy_host' => nil,
+    'proxy_port' => nil,
+  }
+
+  include TestNetHTTPUtils
+
+  def test_partial_response
+    str = "0123456789"
+    @server.mount_proc('/') do |req, res|
+      res.status = 200
+      res['Content-Type'] = 'text/plain'
+
+      res.body = str
+      res['Content-Length'] = str.length + 1
+    end
+    @server.mount_proc('/show_ip') { |req, res| res.body = req.remote_ip }
+
+    http = Net::HTTP.new(config('host'), config('port'))
+    res = http.get('/')
+    assert_equal(str, res.body)
+
+    http = Net::HTTP.new(config('host'), config('port'))
+    http.ignore_eof = false
+    assert_raise(EOFError) {http.get('/')}
+  end
+end
